@@ -1,5 +1,5 @@
 from functools import partial
-from typing import Callable, List, Optional, Sequence, Tuple, Union
+from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union
 
 import torch
 import torch.nn as nn
@@ -60,7 +60,10 @@ class MobileNetV5MultiScaleFusionAdapter(nn.Module):
         noskip: bool = True,
         act_layer: Optional[LayerType] = None,
         norm_layer: Optional[LayerType] = None,
+        device=None,
+        dtype=None,
   ):
+    dd = {'device': device, 'dtype': dtype}
     super().__init__()
     self.in_channels = sum(in_chs) if isinstance(in_chs, Sequence) else in_chs
     self.out_channels = out_chs
@@ -81,9 +84,10 @@ class MobileNetV5MultiScaleFusionAdapter(nn.Module):
         norm_layer=norm_layer,
         noskip=self.noskip,
         layer_scale_init_value=self.layer_scale_init_value,
+        **dd,
     )
 
-    self.norm = norm_layer(self.out_channels)
+    self.norm = norm_layer(self.out_channels, **dd)
 
   def forward(self, inputs: List[torch.Tensor]) -> torch.Tensor:
     # Inputs list of [B, C, H, W] tensors
@@ -129,12 +133,12 @@ class MobileNetV5(nn.Module):
             num_classes: int = 1000,
             in_chans: int = 3,
             stem_size: int = 16,
-            stem_bias: bool = False,
+            stem_bias: bool = True,
             fix_stem: bool = False,
             num_features: int = 2048,
             pad_type: str = '',
             use_msfa: bool = True,
-            msfa_indices: List[int] = (-3, -2, -1),
+            msfa_indices: List[int] = (-2, -1),
             msfa_output_resolution: int = 16,
             act_layer: Optional[LayerType] = None,
             norm_layer: Optional[LayerType] = None,
@@ -146,6 +150,8 @@ class MobileNetV5(nn.Module):
             drop_path_rate: float = 0.,
             layer_scale_init_value: Optional[float] = None,
             global_pool: str = 'avg',
+            device=None,
+            dtype=None,
     ):
         """
         Args:
@@ -169,6 +175,7 @@ class MobileNetV5(nn.Module):
             global_pool: Type of pooling to use for global pooling features of the FC head.
         """
         super().__init__()
+        dd = {'device': device, 'dtype': dtype}
         act_layer = act_layer or _GELU
         norm_layer = get_norm_layer(norm_layer) or RmsNorm2d
         norm_act_layer = get_norm_act_layer(norm_layer, act_layer)
@@ -191,6 +198,7 @@ class MobileNetV5(nn.Module):
             bias=stem_bias,
             norm_layer=norm_layer,
             act_layer=act_layer,
+            **dd,
         )
 
         # Middle stages (IR/ER/DS Blocks)
@@ -205,6 +213,7 @@ class MobileNetV5(nn.Module):
             se_layer=se_layer,
             drop_path_rate=drop_path_rate,
             layer_scale_init_value=layer_scale_init_value,
+            **dd,
         )
         self.blocks = nn.Sequential(*builder(stem_size, block_args))
         self.feature_info = builder.features
@@ -224,6 +233,7 @@ class MobileNetV5(nn.Module):
                 output_resolution=self.msfa_output_resolution,
                 norm_layer=norm_layer,
                 act_layer=act_layer,
+                **dd,
             )
             self.global_pool = SelectAdaptivePool2d(pool_type=global_pool)
             self.conv_head = None
@@ -235,11 +245,11 @@ class MobileNetV5(nn.Module):
             self.global_pool = SelectAdaptivePool2d(pool_type=global_pool)
             num_pooled_chs = self.num_features * self.global_pool.feat_mult()
             # mobilenet-v4 style post-pooling PW conv is followed by a norm+act layer
-            self.conv_head = create_conv2d(num_pooled_chs, self.head_hidden_size, 1, padding=pad_type)
-            self.norm_head = norm_act_layer(self.head_hidden_size)
+            self.conv_head = create_conv2d(num_pooled_chs, self.head_hidden_size, 1, padding=pad_type, **dd)
+            self.norm_head = norm_act_layer(self.head_hidden_size, **dd)
 
         self.flatten = nn.Flatten(1) if global_pool else nn.Identity()  # don't flatten if pooling disabled
-        self.classifier = Linear(self.head_hidden_size, num_classes) if num_classes > 0 else nn.Identity()
+        self.classifier = Linear(self.head_hidden_size, num_classes, **dd) if num_classes > 0 else nn.Identity()
 
         efficientnet_init_weights(self)
 
@@ -426,8 +436,11 @@ class MobileNetV5Encoder(nn.Module):
             drop_rate: float = 0.,
             drop_path_rate: float = 0.,
             layer_scale_init_value: Optional[float] = None,
+            device=None,
+            dtype=None,
     ):
         super().__init__()
+        dd = {'device': device, 'dtype': dtype}
         act_layer = act_layer or _GELU
         norm_layer = get_norm_layer(norm_layer) or RmsNorm2d
         se_layer = se_layer or SqueezeExcite
@@ -447,6 +460,7 @@ class MobileNetV5Encoder(nn.Module):
             bias=stem_bias,
             norm_layer=norm_layer,
             act_layer=act_layer,
+            **dd,
         )
 
         builder = EfficientNetBuilder(
@@ -460,6 +474,7 @@ class MobileNetV5Encoder(nn.Module):
             se_layer=se_layer,
             drop_path_rate=drop_path_rate,
             layer_scale_init_value=layer_scale_init_value,
+            **dd,
         )
         self.blocks = nn.Sequential(*builder(stem_size, block_args))
         self.feature_info = builder.features
@@ -477,6 +492,7 @@ class MobileNetV5Encoder(nn.Module):
             output_resolution=self.msfa_output_resolution,
             norm_layer=norm_layer,
             act_layer=act_layer,
+            **dd,
         )
 
         efficientnet_init_weights(self)
@@ -574,6 +590,19 @@ class MobileNetV5Encoder(nn.Module):
         return self.forward_features(x)
 
 
+def checkpoint_filter_fn(
+        state_dict: Dict[str, torch.Tensor],
+        model,
+) -> Dict[str, torch.Tensor]:
+    """ convert weights from gemma encoders """
+    state_dict = state_dict.get('model', state_dict)
+    state_dict = state_dict.get('state_dict', state_dict)
+    if 'model.vision_tower.timm_model.conv_stem.conv.weight' in state_dict:
+        prefix = 'model.vision_tower.timm_model.'
+        state_dict = {k.replace(prefix, ''): v for k, v in state_dict.items() if prefix in k}
+    return state_dict
+
+
 def _create_mnv5_encoder(variant: str, pretrained: bool = False, **kwargs) -> MobileNetV5Encoder:
     out_indices = kwargs.pop('out_indices', (0, 1, 2, 3, 4))
     feature_cfg = dict(out_indices=out_indices, feature_cls='getter')
@@ -590,6 +619,7 @@ def _create_mnv5_encoder(variant: str, pretrained: bool = False, **kwargs) -> Mo
         variant,
         pretrained,
         pretrained_strict=False,
+        pretrained_filter_fn=checkpoint_filter_fn,
         feature_cfg=feature_cfg,
         kwargs_filter=kwargs_filter,
         **kwargs,
@@ -597,14 +627,14 @@ def _create_mnv5_encoder(variant: str, pretrained: bool = False, **kwargs) -> Mo
     return model
 
 
-def _create_mnv5(variant: str, pretrained: bool = False, **kwargs) -> MobileNetV5Encoder:
+def _create_mnv5(variant: str, pretrained: bool = False, **kwargs) -> MobileNetV5:
     out_indices = kwargs.pop('out_indices', (0, 1, 2, 3, 4))
     feature_cfg = dict(out_indices=out_indices, feature_cls='getter')
     model = build_model_with_cfg(
         MobileNetV5,
         variant,
         pretrained,
-        pretrained_strict=False,
+        pretrained_filter_fn=checkpoint_filter_fn,
         feature_cfg=feature_cfg,
         **kwargs,
     )
@@ -801,19 +831,21 @@ def _cfg(url: str = '', **kwargs):
 
 
 default_cfgs = generate_default_cfgs({
-    # encoder-only configs
+    # Encoder-only config for Gemma 3n Transformers integration
     'mobilenetv5_300m_enc': _cfg(
-        #hf_hub_id='timm/',
         mean=(0., 0., 0.), std=(1., 1., 1.),
         input_size=(3, 768, 768),
         num_classes=0),
 
-    # WIP classification configs for testing
-    'mobilenetv5_300m': _cfg(
-        # hf_hub_id='timm/',
+    # Gemma 3n encoder weights for timm use / fine-tune
+    'mobilenetv5_300m.gemma3n': _cfg(
+        hf_hub_id='timm/',
         mean=(0., 0., 0.), std=(1., 1., 1.),
         input_size=(3, 768, 768),
-        num_classes=0),
+        num_classes=0,
+        license='gemma'),
+
+    # WIP classification configs for testing
     'mobilenetv5_base.untrained': _cfg(
         # hf_hub_id='timm/',
         num_classes=1000)
