@@ -132,6 +132,7 @@ class NaFlexVitCfg:
     act_layer: Optional[str] = None  # Activation layer for MLP blocks
     block_fn: Optional[str] = None  # Transformer block implementation class name
     mlp_layer: Optional[str] = None  # MLP implementation class name
+    attn_layer: Optional[str] = None  # Attention layer implementation (e.g., 'attn', 'diff')
 
     # EVA-specific parameters
     attn_type: str = 'standard'  # Attention type: 'standard', 'eva', 'rope'
@@ -289,13 +290,15 @@ def get_block_fn(cfg: NaFlexVitCfg) -> Callable:
     else:
         # Standard ViT block
         block_fn = cfg.block_fn or Block
+        block_kwargs = {}
         if cfg.scale_mlp_norm or cfg.scale_attn_inner_norm:
             # param names differ between EVA vs non-EVA block types
-            block_fn = partial(
-                block_fn,
-                scale_mlp_norm=cfg.scale_mlp_norm,
-                scale_attn_norm=cfg.scale_attn_inner_norm
-            )
+            block_kwargs['scale_mlp_norm'] = cfg.scale_mlp_norm
+            block_kwargs['scale_attn_norm'] = cfg.scale_attn_inner_norm
+        if cfg.attn_layer:
+            block_kwargs['attn_layer'] = cfg.attn_layer
+        if block_kwargs:
+            block_fn = partial(block_fn, **block_kwargs)
         return block_fn
 
 
@@ -405,8 +408,8 @@ class NaFlexEmbeds(nn.Module):
         self.num_prefix_tokens += reg_tokens
 
         # Create class and register tokens
-        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim, **dd)) if class_token else None
-        self.reg_token = nn.Parameter(torch.zeros(1, reg_tokens, embed_dim, **dd)) if reg_tokens else None
+        self.cls_token = nn.Parameter(torch.empty(1, 1, embed_dim, **dd)) if class_token else None
+        self.reg_token = nn.Parameter(torch.empty(1, reg_tokens, embed_dim, **dd)) if reg_tokens else None
 
         # Calculate grid size and number of patches
         self.default_img_size: Optional[Tuple[int, int]] = None
@@ -478,17 +481,32 @@ class NaFlexEmbeds(nn.Module):
             assert self.pos_embed_grid_size is not None
             h, w = self.pos_embed_grid_size
             self.pos_embed_type = 'factorized'
-            self.pos_embed_y = nn.Parameter(torch.randn(1, h, embed_dim, **dd) * .02)
-            self.pos_embed_x = nn.Parameter(torch.randn(1, w, embed_dim, **dd) * .02)
+            self.pos_embed_y = nn.Parameter(torch.empty(1, h, embed_dim, **dd))
+            self.pos_embed_x = nn.Parameter(torch.empty(1, w, embed_dim, **dd))
         else:
             assert self.pos_embed_grid_size is not None
             h, w = self.pos_embed_grid_size
-            self.pos_embed = nn.Parameter(torch.randn(1, h, w, embed_dim, **dd) * .02)
+            self.pos_embed = nn.Parameter(torch.empty(1, h, w, embed_dim, **dd))
             self.pos_embed_type = 'learned'
 
         # Dropout layer
         self.pos_drop = nn.Dropout(p=pos_drop_rate)
 
+        if not self.proj.weight.is_meta:
+            self.reset_parameters()
+
+    def reset_parameters(self) -> None:
+        if self.cls_token is not None:
+            nn.init.normal_(self.cls_token, std=1e-6)
+        if self.reg_token is not None:
+            nn.init.normal_(self.reg_token, std=1e-6)
+        if self.pos_embed is not None:
+            nn.init.normal_(self.pos_embed, std=.02)
+        if self.pos_embed_y is not None:
+            nn.init.normal_(self.pos_embed_y, std=.02)
+        if self.pos_embed_x is not None:
+            nn.init.normal_(self.pos_embed_x, std=.02)
+
     def feature_info(self, location) -> Dict[str, Any]:
         """Get feature information for feature extraction.
 
@@ -1214,6 +1232,7 @@ class NaFlexVit(nn.Module):
                 norm_layer=norm_layer,
                 act_layer=act_layer,
                 mlp_layer=mlp_layer,
+                depth=i,
                 **dd,
             )
             for i in range(cfg.depth)
@@ -1249,10 +1268,10 @@ class NaFlexVit(nn.Module):
         self.head_drop = nn.Dropout(cfg.drop_rate)
         self.head = nn.Linear(self.embed_dim, num_classes, **dd) if num_classes > 0 else nn.Identity()
 
-        if cfg.weight_init != 'skip':
-            self.init_weights(cfg.weight_init)
-        if cfg.fix_init:
-            self.fix_init_weight()
+        self.weight_init_mode = cfg.weight_init
+        self.fix_init = cfg.fix_init
+        if not self.embeds.proj.weight.is_meta:
+            self.init_weights(cfg.weight_init, needs_reset=False)
 
     def fix_init_weight(self) -> None:
         """Apply initialization weight fix with layer-wise scaling."""
@@ -1271,15 +1290,22 @@ class NaFlexVit(nn.Module):
                 rescale(layer.mlp_out_proj.weight, layer_id + 1)
 
 
-    def init_weights(self, mode: str = '') -> None:
+    def init_weights(self, mode: str = '', needs_reset: bool = True) -> None:
         """Initialize model weights according to specified scheme.
 
         Args:
             mode: Initialization mode ('jax', 'jax_nlhb', 'moco', or '')
+            needs_reset: If True, call reset_parameters() on modules (default for after to_empty()).
+                If False, skip reset_parameters() (for __init__ where modules already self-initialized).
         """
+        mode = mode or self.weight_init_mode
         assert mode in ('jax', 'jax_nlhb', 'moco', '')
         head_bias = -math.log(self.num_classes) if 'nlhb' in mode else 0.
-        named_apply(get_init_weights_vit(mode, head_bias), self)
+
+        named_apply(get_init_weights_vit(mode, head_bias, needs_reset=needs_reset), self)
+
+        if self.fix_init:
+            self.fix_init_weight()
 
     @torch.jit.ignore()
     def load_pretrained(self, checkpoint_path: str, prefix: str = '') -> None:
@@ -1822,16 +1848,21 @@ def _debug_dump_patches(x):
         save_image(patch, f'patch_{i}.jpg', normalize=True)
 
 
-def get_init_weights_vit(mode: str = 'jax', head_bias: float = 0.0) -> Callable:
+def get_init_weights_vit(mode: str = 'jax', head_bias: float = 0.0, needs_reset: bool = True) -> Callable:
     """Function imported from vision_transformer.py to maintain compatibility"""
-    from .vision_transformer import init_weights_vit_jax, init_weights_vit_moco, init_weights_vit_timm
+    from .vision_transformer import (
+        init_weights_vit_jax,
+        init_weights_vit_moco,
+        init_weights_vit_timm,
+        init_weights_reset_parameters,
+    )
 
     if 'jax' in mode:
-        return partial(init_weights_vit_jax, head_bias=head_bias)
+        return partial(init_weights_vit_jax, head_bias=head_bias, needs_reset=needs_reset)
     elif 'moco' in mode:
-        return init_weights_vit_moco
+        return partial(init_weights_vit_moco, needs_reset=needs_reset)
     else:
-        return init_weights_vit_timm
+        return partial(init_weights_vit_timm, needs_reset=needs_reset)
 
 
 def checkpoint_filter_fn(state_dict: Dict[str, Any], model: NaFlexVit) -> Dict[str, Any]:
