@@ -13,6 +13,7 @@ Modifications and additions for timm hacked together by / Copyright 2022, Ross W
 # Written by Ze Liu
 # --------------------------------------------------------
 import math
+from functools import partial
 from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union
 
 import torch
@@ -115,7 +116,7 @@ class WindowAttention(nn.Module):
         self.num_heads = num_heads
         self.qkv_bias_separate = qkv_bias_separate
 
-        self.logit_scale = nn.Parameter(torch.log(10 * torch.ones((num_heads, 1, 1), **dd)))
+        self.logit_scale = nn.Parameter(torch.empty((num_heads, 1, 1), **dd))
 
         # mlp to generate continuous relative position bias
         self.cpb_mlp = nn.Sequential(
@@ -126,9 +127,9 @@ class WindowAttention(nn.Module):
 
         self.qkv = nn.Linear(dim, dim * 3, bias=False, **dd)
         if qkv_bias:
-            self.q_bias = nn.Parameter(torch.zeros(dim, **dd))
-            self.register_buffer('k_bias', torch.zeros(dim, **dd), persistent=False)
-            self.v_bias = nn.Parameter(torch.zeros(dim, **dd))
+            self.q_bias = nn.Parameter(torch.empty(dim, **dd))
+            self.register_buffer('k_bias', torch.empty(dim, **dd), persistent=False)
+            self.v_bias = nn.Parameter(torch.empty(dim, **dd))
         else:
             self.q_bias = None
             self.k_bias = None
@@ -138,10 +139,50 @@ class WindowAttention(nn.Module):
         self.proj_drop = nn.Dropout(proj_drop)
         self.softmax = nn.Softmax(dim=-1)
 
-        self._make_pair_wise_relative_positions(**dd)
+        # Register empty buffers with correct shapes
+        win_h, win_w = self.window_size
+        self.register_buffer(
+            "relative_coords_table",
+            torch.empty(1, 2 * win_h - 1, 2 * win_w - 1, 2, **dd),
+            persistent=False,
+        )
+        self.register_buffer(
+            "relative_position_index",
+            torch.empty(win_h * win_w, win_h * win_w, device=device, dtype=torch.long),
+            persistent=False,
+        )
 
-    def _make_pair_wise_relative_positions(self, device=None, dtype=None) -> None:
-        """Create pair-wise relative position index and coordinates table."""
+        if not self.proj.weight.is_meta:
+            self.reset_parameters()
+
+    def reset_parameters(self) -> None:
+        """Initialize parameters and buffers."""
+        nn.init.constant_(self.logit_scale, math.log(10))
+        if self.q_bias is not None:
+            nn.init.zeros_(self.q_bias)
+            nn.init.zeros_(self.v_bias)
+        self._init_buffers()
+
+    def _init_buffers(self) -> None:
+        """Compute and fill non-persistent buffer values."""
+        if self.k_bias is not None:
+            self.k_bias.zero_()
+        relative_coords_table, relative_position_index = self._make_pair_wise_relative_positions(
+            device=self.proj.weight.device, dtype=self.proj.weight.dtype
+        )
+        self.relative_coords_table.copy_(relative_coords_table)
+        self.relative_position_index.copy_(relative_position_index)
+
+    def _make_pair_wise_relative_positions(
+            self,
+            device=None,
+            dtype=None,
+    ) -> Tuple[torch.Tensor, torch.Tensor]:
+        """Compute pair-wise relative position index and coordinates table.
+
+        Returns:
+            Tuple of (relative_coords_table, relative_position_index)
+        """
         # get relative_coords_table
         relative_coords_h = torch.arange(
             -(self.window_size[0] - 1), self.window_size[0], device=device, dtype=torch.float32)
@@ -158,7 +199,7 @@ class WindowAttention(nn.Module):
         relative_coords_table *= 8  # normalize to -8, 8
         relative_coords_table = torch.sign(relative_coords_table) * torch.log2(
             torch.abs(relative_coords_table) + 1.0) / math.log2(8)
-        self.register_buffer("relative_coords_table", relative_coords_table.to(dtype=dtype), persistent=False)
+        relative_coords_table = relative_coords_table.to(dtype=dtype)
 
         # get pair-wise relative position index for each token inside the window
         coords_h = torch.arange(self.window_size[0], device=device, dtype=torch.long)
@@ -171,7 +212,8 @@ class WindowAttention(nn.Module):
         relative_coords[:, :, 1] += self.window_size[1] - 1
         relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
         relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
-        self.register_buffer("relative_position_index", relative_position_index, persistent=False)
+
+        return relative_coords_table, relative_position_index
 
     def set_window_size(self, window_size: Tuple[int, int]) -> None:
         """Update window size and regenerate relative position tables.
@@ -185,7 +227,14 @@ class WindowAttention(nn.Module):
             device = self.relative_coords_table.device
             dtype = self.relative_coords_table.dtype
             self.window_size = window_size
-            self._make_pair_wise_relative_positions(device=device, dtype=dtype)
+            relative_coords_table, relative_position_index = \
+                self._make_pair_wise_relative_positions(device=device, dtype=dtype)
+            self.register_buffer("relative_coords_table", relative_coords_table, persistent=False)
+            self.register_buffer("relative_position_index", relative_position_index, persistent=False)
+
+    def init_non_persistent_buffers(self) -> None:
+        """Initialize non-persistent buffers."""
+        self._init_buffers()
 
     def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
         """Forward pass of window attention.
@@ -795,20 +844,33 @@ class SwinTransformerV2(nn.Module):
             **dd,
         )
 
-        self.apply(self._init_weights)
-        for bly in self.layers:
-            bly._init_respostnorm()
+        if not self.patch_embed.proj.weight.is_meta:
+            self.init_weights(needs_reset=False)
 
-    def _init_weights(self, m: nn.Module) -> None:
+    def init_weights(self, needs_reset: bool = True) -> None:
         """Initialize model weights.
 
+        Args:
+            needs_reset: If True, call reset_parameters() on modules (default for after to_empty()).
+                If False, skip reset_parameters() (for __init__ where modules already self-initialized).
+        """
+        self.apply(partial(self._init_weights, needs_reset=needs_reset))
+        for bly in self.layers:
+            bly._init_respostnorm()
+
+    def _init_weights(self, m: nn.Module, needs_reset: bool = True) -> None:
+        """Initialize weights for Linear layers.
+
         Args:
             m: Module to initialize.
+            needs_reset: Whether to call reset_parameters() on modules.
         """
         if isinstance(m, nn.Linear):
             trunc_normal_(m.weight, std=.02)
-            if isinstance(m, nn.Linear) and m.bias is not None:
+            if m.bias is not None:
                 nn.init.constant_(m.bias, 0)
+        elif needs_reset and hasattr(m, 'reset_parameters'):
+            m.reset_parameters()
 
     def set_input_size(
             self,
